<?php
namespace WCStepFilter;

/**
 * Core Class
 *
 * @class Core
 * @version 4.5.2
 */
class Core
{
    // <editor-fold desc="Properties">
    /**
     * Plugin global settings model
     *
     * @var array
     */
    public static $settingsModel;

    /**
     * Array of the showed templates
     * @var array
     */
    public static $templatesShowed = [];

    /**
     * Filter class instance
     *
     * @var object
     */
    public static $filter;

    /**
     * Question class instance
     *
     * @var object
     */
    public static $question;

    /**
     * Inquiry class instance
     *
     * @var object
     */
    public static $inquiry;

    /**
     * Product attribute value settings model
     *
     * @var object
     */
    public static $attributeValueSettingsModel;

    /**
     * Product category settings model
     *
     * @var object
     */
    public static $categorySettingsModel;
    // </editor-fold>

    // <editor-fold desc="Core">
    /** Constructor */
    public function __construct()
    {
        if (!class_exists('\\WCStepFilter\\Filter')) {
            include_once('Filter.php');
        }

        if (!class_exists('\\WCStepFilter\\Question')) {
            include_once('Question.php');
        }

        if (!class_exists('\\WCStepFilter\\Inquiry')) {
            include_once('Inquiry.php');
        }

        // actions
        add_action('init', [$this, 'onInit']);
        add_action('woocommerce_init', [$this, 'onWoocommerceInit'], 1);
        add_action('wp_loaded', [$this, 'requests']);
        add_action('woocommerce_before_shop_loop', [$this, 'outputShopControls']);
        add_action('woocommerce_no_products_found', [$this, 'outputShopControls']);

        if (isset($_GET['wcsf-results'])) {
            add_action('woocommerce_product_query', [$this, 'modifyProductsQuery'], 15);

            // filtering WOOF plugin args
            add_filter('woof_products_query', [$this, 'modifyWoofQuery'], 15);
        }

        // settings page
        add_filter('woocommerce_settings_tabs_array', [$this, 'addSettingsTab'], 50);
        add_action('woocommerce_settings_tabs_step_filter_settings_tab', [$this, 'initSettingsTab']);
        add_action('woocommerce_update_options_step_filter_settings_tab', [$this, 'updateSettingsTab']);

        // front scripts and styles
        add_action('wp_enqueue_scripts', [$this, 'enqueueFront']);

        // admin scripts and styles
        add_action('admin_enqueue_scripts', [$this, 'enqueueAdmin']);

        // init basic classes
        self::$inquiry = new Inquiry();
        self::$filter = new Filter();
        self::$question = new Question();
    }

    /** On init action */
    public function onInit()
    {
        self::$settingsModel = apply_filters(
            'wcsf_global_settings_model',
            [
                'include_full_style' => [
                    'name' => esc_html__('Include full styles file', 'woocommerce-step-filter'),
                    'default' => 'yes',
                    'type' => 'checkbox',
                    'id' => 'woocommerce_step_filter_include_full_styles_file',
                    'desc' => 'If the active theme doesn\'t based on "Bootstrap" framework'
                ],
                'scripts_including_type' => [
                    'name' => esc_html__('Scripts including type', 'woocommerce-step-filter'),
                    'default' => 'single',
                    'type' => 'select',
                    'desc' => 'Include scripts separately or in a single file',
                    'id' => 'woocommerce_step_filter_scripts_including_type',
                    'options' => [
                        'single' => esc_html__('Single file', 'woocommerce-step-filter'),
                        'multiple' => esc_html__('Multiple files', 'woocommerce-step-filter')
                    ]
                ],
                'included_scripts' => [
                    'name' => esc_html__('Included scripts', 'woocommerce-step-filter'),
                    'default' => [
                        'app',
                        'hooks',
                        'jquery.serializejson',
                        'wNumb',
                        'nouislider',
                        'nouislider-launcher'
                    ],
                    'type' => 'multiselect',
                    'desc' => 'Select which files will be included with the multiple including type',
                    'id' => 'woocommerce_step_filter_included_scripts',
                    'options' => [
                        'app' => 'app',
                        'hooks' => 'hooks',
                        'jquery.serializejson' => 'jquery.serializejson',
                        'wNumb' => 'wNumb',
                        'nouislider' => 'nouislider',
                        'nouislider-launcher' => 'nouislider-launcher'
                    ]
                ]
            ]
        );

        self::$attributeValueSettingsModel = apply_filters(
            'wcsf_attribute_value_settings_model',
            [
                'pre-checked' => [
                    'label' => esc_html__('Step Filter Pre-checked', 'woocommerce-step-filter'),
                    'key' => '_wcsf_pre_checked',
                    'type' => 'checkbox',
                    'default' => false
                ],
                'thumbnail' => [
                    'label' => esc_html__('Step Filter Thumbnail', 'woocommerce-step-filter'),
                    'key' => '_wcsf_thumbnail',
                    'type' => 'thumbnail',
                    'default' => ''
                ]
            ]
        );

        self::$categorySettingsModel = apply_filters(
            'wcsf_category_settings_model',
            [
                'pre-checked' => [
                    'label' => esc_html__('Step Filter Pre-checked', 'woocommerce-step-filter'),
                    'key' => '_wcsf_pre_checked',
                    'type' => 'checkbox',
                    'default' => false
                ]
            ]
        );
    }

    /** Runs after woocommerce is loaded */
    public function onWoocommerceInit()
    {
        if (!session_id() && !(is_admin() && !is_ajax())) {
            session_start();
        }

        add_action('wcsf_shortcode_init', [$this, 'shortCodeAction']);
        add_action('product_cat_edit_form_fields', [$this, 'categoryFieldsAdd'], 10, 2);
        add_action('product_tag_edit_form_fields', [$this, 'categoryFieldsAdd'], 10, 2);
        add_action('edited_product_cat', [$this, 'categoryFieldsSave'], 10, 2);
        add_action('edited_product_tag', [$this, 'categoryFieldsSave'], 10, 2);

        $attributes = \wc_get_attribute_taxonomies();

        foreach ($attributes as $attribute) {
            add_action(
                'pa_' . $attribute->attribute_name . '_edit_form_fields',
                [$this, 'attributeValueFieldsAdd'],
                10,
                2
            );

            add_action('edited_pa_' . $attribute->attribute_name, [$this, 'attributeValueFieldsSave'], 10, 2);
        }
    }

    /**
     * Short-code call action
     *
     * @param array $attributes
     */
    public function shortCodeAction($attributes)
    {
        if (!isset($attributes['id']) || empty($attributes['id'])) {
            return;
        }

        $id = (int) $attributes['id'];
        $resultsUrl = self::$filter->getSetting($id, 'results_url');

        if ($resultsUrl == '#') {
            if (!isset($_GET['wcsf-results']) || empty($_GET['wcsf-results'])) {
                // reset filter's state if it's results page is any and there is no query parameters
                self::$filter->resetState($id);
            } elseif (isset($_GET['wcsf-value']) && !empty($_GET['wcsf-value']) && $_GET['wcsf-results'] == $id) {
                // set filter's value from get parameter
                self::$filter->setValue($id, $_GET['wcsf-value']);
            }
        }
    }
    // </editor-fold>

    // <editor-fold desc="HTML">
    /** Output shop controls html */
    public function outputShopControls()
    {
        static $wcsfShopControlsShown;

        if (isset($_GET['wcsf-results']) && !$wcsfShopControlsShown) {
            $wcsfShopControlsShown = true;

            self::getTemplatePart('shop-controls/index', ['id' => (int) $_GET['wcsf-results']]);
        }
    }

    /**
     * Include php-template by the name.
     * First looking in the "theme folder/woocommerce-step-filter (WC_STEP_FILTER_THEME_TEMPLATES_DIR)"
     * Second looking in the "plugin folder/views"
     * Making extraction of the arguments as variables
     *
     * @param string $name
     * @param array  $arguments
     * @param array  $settings
     *
     * @return string
     */
    public static function getTemplatePart($name = '', $arguments = [], $settings = [])
    {
        $defaultSettings = [
            'echo' => true,
            'once' => false,
            'compress' => !WC_STEP_FILTER_DEBUG
        ];

        $settings = array_merge($defaultSettings, $settings);

        // show template only once
        if ($settings['once'] && in_array($name, self::$templatesShowed)) {
            return null;
        }

        // save template as showed
        self::$templatesShowed[] = $name;

        if (is_array($arguments)) {
            extract($arguments, EXTR_PREFIX_SAME, 'data');
        }

        $path = get_stylesheet_directory() . DIRECTORY_SEPARATOR
            . WC_STEP_FILTER_THEME_TEMPLATES_DIR . DIRECTORY_SEPARATOR . $name . '.php';

        if (!file_exists($path)) {
            $path = WC_STEP_FILTER_PLUGIN_PATH . DIRECTORY_SEPARATOR
                . 'views' . DIRECTORY_SEPARATOR . $name . '.php';
        }

        if (!file_exists($path)) {
            return '';
        }

        ob_start();

        include($path);

        $html = ob_get_clean();

        if ($settings['compress']) {
            $replace = ['>', '<', '\\1', ' ', ''];
            $search = [
                '/\>[^\S ]+/s',     // strip whitespaces after tags, except space
                '/[^\S ]+\</s',     // strip whitespaces before tags, except space
                '/(\s)+/s',         // shorten multiple whitespace sequences
                '/<!--spacer-->/',  // replace spacer tag
                '/<!--(.|\s)*?-->/' // remove HTML comments
            ];

            $html = preg_replace($search, $replace, $html);
        }

        if ($settings['echo']) {
            echo $html;

            return null;
        }

        return $html;
    }
    // </editor-fold>

    // <editor-fold desc="Query">
    /** Add request actions */
    public function requests()
    {
        if (is_admin()) {
            return;
        }

        // get queries
        // reset the filter
        if (isset($_GET['wcsf-reset-state'])) {
            self::$filter->resetState((int) $_GET['wcsf-reset-state']);
        }

        // reset one value
        if (isset($_GET['wcsf-reset-value'], $_GET['wcsf-results'])) {
            self::$filter->resetValue((int) $_GET['wcsf-results'], (int) $_GET['wcsf-reset-value']);
        }

        // post queries
        if (!isset($_POST['wcsf'])) {
            return;
        }

        $id = isset($_POST['id']) ? (int) $_POST['id'] : false;

        // save value
        if (isset($_POST['value']) && !empty($_POST['value'])) {
            self::$filter->setValue($id, (array) $_POST['value']);
        }

        // change step
        if (isset($_POST['stepId'])) {
            self::$filter->setActiveStep($id, (int) $_POST['stepId']);
        }

        // reset filter
        if (isset($_POST['reset'])) {
            self::$filter->resetState($id);
        }

        // show results page
        if (isset($_POST['results'])) {
            $redirectLink = self::$filter->getResultsUrl($id);

            do_action('wcsf_before_results_redirect', $id, $redirectLink);

            // redirect to the shop page
            wp_redirect($redirectLink, 200);

            // js version of redirect
            exit("<script>document.location = '$redirectLink';</script>");
        }
    }

    /**
     * Get query args by the filter id and value
     *
     * @param integer $filterId
     * @param array $value
     * @param array $args
     *
     * @return array
     */
    public static function getQueryArgsByFilter($filterId, $value, $args = [])
    {
        $questions = self::$question->getList(self::$filter->getSetting($filterId, 'questions'));
        $maxResultsNumber = self::$filter->getSetting($filterId, 'max_results_number');
        $taxQuery = [];
        $postIn = [];
        $metaQuery = [];
        $multiChoiceOperators = apply_filters(
            'wsf_multi_choice_operators',
            [
                -1 => 'NOT IN',
                1 => 'IN',
                2 => 'AND'
            ]
        );

        // blend in passed category or tag id
        if (isset($args['categoryId']) && $args['categoryId']) {
            $taxQuery[] = [
                'taxonomy' => 'product_cat',
                'field' => 'id',
                'terms' => $args['categoryId']
            ];
        } elseif (isset($args['tagId']) && $args['tagId']) {
            $taxQuery[] = [
                'taxonomy' => 'product_tag',
                'field' => 'id',
                'terms' => $args['tagId']
            ];
        }

        // loop for the each filter value
        foreach ((array) $value as $questionId => $questionValue) {
            // question doesn't exist
            if (!isset($questions[$questionId])) {
                continue;
            }

            $question = $questions[$questionId];
            $viewType = $question->viewType;
            $filterType = $question->filterType;
            $relation = isset($question->relation) ? $question->relation : 'or';
            $operator = $relation == 'or' ? 'IN' : 'AND';

            if (in_array($filterType, ['tag', 'category', 'attribute'])) {
                // attribute, category or tag
                if ($filterType == 'attribute') {
                    // filtering by attribute
                    if (!isset($question->attribute)) {
                        continue;
                    }

                    $taxonomy = \wc_attribute_taxonomy_name_by_id((int) $question->attribute);
                } else {
                    // filtering by tag or category
                    $taxonomy = $filterType == 'tag' ? 'product_tag' : 'product_cat';
                }

                if (strpos($viewType, 'number') !== false) {
                    // numeric type of field
                    $filterBy = [];
                    $terms = get_terms(['taxonomy' => $taxonomy]);

                    foreach ($terms as $term) {
                        $termValue = (float) $term->name;
                        $issetFrom = isset($questionValue['from']) && !empty($questionValue['from']);
                        $from = $issetFrom && $questionValue['from'] <= $termValue;
                        $issetTo = isset($questionValue['to']) && !empty($questionValue['to']);
                        $to = $issetTo && $questionValue['to'] >= $termValue;

                        // check by numeric with the first, the second and both values
                        if (($viewType == 'number-from' && $from)
                            || ($viewType == 'number-to' && $to)
                            || ($viewType == 'number-between'
                                && ($from && $to || !$issetTo && $from || !$issetFrom && $to))
                        ) {
                            $filterBy[] = $term->term_id;
                        }
                    }

                    if (!empty($filterBy)) {
                        $taxQuery[] = [
                            'taxonomy' => $taxonomy,
                            'field' => 'id',
                            'terms' => $filterBy,
                            'operator' => 'IN'
                        ];
                    }
                } else {
                    if (strpos($viewType, 'multi-choice') !== false) {
                        $in = [];

                        foreach (array_filter($questionValue) as $termId => $choiceValue) {
                            if ($choiceValue == 1) {
                                // get all IN values in the array
                                $in[] = $termId;

                                continue;
                            }

                            $taxQuery[] = [
                                'taxonomy' => $taxonomy,
                                'field' => 'id',
                                'terms' => $termId,
                                'operator' => $multiChoiceOperators[$choiceValue]
                            ];
                        }

                        if (!empty($in)) {
                            $taxQuery[] = [
                                'taxonomy' => $taxonomy,
                                'field' => 'id',
                                'terms' => $in,
                                'operator' => $multiChoiceOperators[1]
                            ];
                        }
                    } elseif (!empty(array_filter($questionValue))) {
                        // skip "doesn't matter" and not-defined value and filter by other terms
                        $taxQuery[] = [
                            'taxonomy' => $taxonomy,
                            'field' => 'id',
                            'terms' => array_filter($questionValue),
                            'operator' => $operator
                        ];
                    }
                }
            } elseif ($filterType == 'manual') {
                // filtering by a manual value
                foreach ($questionValue as $questionValueItem) {
                    if (!isset($question->manualValues[$questionValueItem])) {
                        continue;
                    }

                    $filterBy = $question->manualValues[$questionValueItem]['filter-by'];
                    $items = isset($question->manualValues[$questionValueItem]['items'])
                        ? $question->manualValues[$questionValueItem]['items']
                        : [];

                    $items = !empty($items) ? $items : [-1]; // if nothing is selected

                    if ($filterBy == 'any') {
                        continue;
                    } elseif ($filterBy == 'category') {
                        $taxQuery[] = [
                            'taxonomy' => 'product_cat',
                            'field' => 'id',
                            'terms' => $items,
                            'operator' => $operator
                        ];
                    } elseif ($filterBy == 'product') {
                        if (empty($postIn)) {
                            $postIn = $items;
                        } else {
                            $postIn = array_intersect($postIn, $items);
                        }
                    }
                }
            } elseif (in_array($filterType, ['price', 'meta'])) {
                // filtering by price
                $metaKey = $filterType == 'price' ? '_price' : $question->metaKey;
                $metaValue = null;
                $compare = '=';
                $type = 'CHAR';

                if (strpos($viewType, 'number') !== false) {
                    $type = 'NUMERIC';

                    switch ($viewType) {
                        case 'number-from':
                            $compare = '>=';

                            if (isset($questionValue['from']) && !empty($questionValue['from'])) {
                                $metaValue = $questionValue['from'];
                            }

                            break;
                        case 'number-to':
                            $compare = '<=';

                            if (isset($questionValue['to']) && !empty($questionValue['to'])) {
                                $metaValue = $questionValue['to'];
                            }

                            break;
                        case 'number-between':
                            if (isset($questionValue['from']) && !empty($questionValue['from'])
                                && isset($questionValue['to'])
                                && !empty($questionValue['to'])
                            ) {
                                $compare = 'BETWEEN';
                                $metaValue = $questionValue;
                            } elseif (isset($questionValue['from']) && !empty($questionValue['from'])) {
                                $compare = '>=';
                                $metaValue = $questionValue['from'];
                            } elseif (isset($questionValue['to']) && !empty($questionValue['to'])) {
                                $compare = '<=';
                                $metaValue = $questionValue['to'];
                            }
                    }
                } elseif (strpos($viewType, 'multi-choice') !== false) {
                    $in = [];

                    foreach (array_filter($questionValue) as $metaValue => $choiceValue) {
                        if ($choiceValue == 1) {
                            // get all IN values in the array
                            $in[] = $metaValue;

                            continue;
                        }

                        $metaQuery[] = [
                            'key' => $metaKey,
                            'value' => $metaValue,
                            'compare' => $multiChoiceOperators[$choiceValue],
                            'type' => $type
                        ];
                    }

                    if (!empty($in)) {
                        $metaQuery[] = [
                            'key' => $metaKey,
                            'value' => $in,
                            'compare' => $multiChoiceOperators[1],
                            'type' => $type
                        ];
                    }

                    $metaValue = null;
                } else {
                    $metaValue = $questionValue;
                }

                if (!is_null($metaValue)) {
                    $metaQuery[] = [
                        'key' => $metaKey,
                        'value' => $metaValue,
                        'compare' => $compare,
                        'type' => $type
                    ];
                }
            }
        }

        return apply_filters(
            'wcsf_query_args_by_filter',
            [
                'tax_query' => $taxQuery,
                'meta_query' => $metaQuery,
                'post__in' => $postIn,
                'maxResultsNumber' => $maxResultsNumber
            ]
        );
    }

    /**
     * Filter the products on the shop page
     *
     * @param array $query
     *
     * @return null
     */
    public function modifyProductsQuery($query)
    {
        if (is_admin() || !isset($_GET['wcsf-results']) || !isset($_GET['wcsf-value'])) {
            return;
        }

        do_action('wcsf_before_modify_request', $query);

        $args = $this->getQueryArgsByFilter((int) $_GET['wcsf-results'], (array) $_GET['wcsf-value']);

        if (method_exists($query, 'set')) {
            if (!empty($args['tax_query'])) {
                $query->set('tax_query', $args['tax_query']);
            }

            if (!empty($args['post__in'])) {
                $query->set('post__in', $args['post__in']);
            }

            if (!empty($args['meta_query'])) {
                $query->set('meta_query', $args['meta_query']);
            }

            // override products number in the query
            if ($args['maxResultsNumber']) {
                $query->set('posts_per_page', $args['maxResultsNumber']);
                $maxResultsNumber = $args['maxResultsNumber'];

                add_filter(
                    'found_posts',
                    function ($found) use ($maxResultsNumber) {
                        return min($found, $maxResultsNumber);
                    },
                    15
                );
            }
        }

        do_action('wcsf_after_modify_request', $query);

        // remove filter
        remove_action('woocommerce_product_query', [$this, 'modifyProductsQuery']);

        return;
    }

    /**
     * Filter the WOOF plugin query
     *
     * @param array $query
     *
     * @return array
     */
    public function modifyWoofQuery($query)
    {
        if (is_admin() || !isset($_GET['wcsf-results']) || !isset($_GET['wcsf-value'])) {
            return $query;
        }

        do_action('wcsf_before_modify_request', $query);

        $args = $this->getQueryArgsByFilter((int) $_GET['wcsf-results'], (array) $_GET['wcsf-value']);
        $query = array_replace($query, $query);

        // override products number in the query
        if ($args['maxResultsNumber']) {
            $query['posts_per_page'] = $args['maxResultsNumber'];
            $maxResultsNumber = $args['maxResultsNumber'];

            add_filter(
                'found_posts',
                function ($args) use ($maxResultsNumber) {
                    return min($args[0], $maxResultsNumber);
                }
            );
        }

        do_action('wcsf_after_modify_request', $query);

        // remove filter
        remove_action('woof_products_query', [$this, 'modifyWoofQuery']);

        return $query;
    }
    // </editor-fold>

    // <editor-fold desc="Assets">
    /** Admin styles and scripts enqueue */
    public function enqueueAdmin()
    {
        $path = WC_STEP_FILTER_DEBUG ? 'src' : 'assets';
        $suffix = WC_STEP_FILTER_DEBUG ? '' : '.min';
        $stylesFolder = WC_STEP_FILTER_DEBUG ? 'scss' : 'css';

        wp_enqueue_media();
        wp_enqueue_script('jquery');
        wp_enqueue_script('jquery-ui-sortable');
        wp_enqueue_script(
            'woocommerce-step-filter-data-table',
            WC_STEP_FILTER_PLUGIN_URL . "$path/admin/js/data-table$suffix.js",
            [
                'jquery',
                'jquery-ui-sortable'
            ],
            WC_STEP_FILTER_VERSION,
            true
        );

        wp_enqueue_script(
            'woocommerce-step-filter-ajax-select',
            WC_STEP_FILTER_PLUGIN_URL . "$path/admin/js/ajax-select$suffix.js",
            ['jquery'],
            WC_STEP_FILTER_VERSION,
            true
        );
        
        wp_enqueue_script(
            'woocommerce-step-filter-thumbnail',
            WC_STEP_FILTER_PLUGIN_URL . "$path/admin/js/thumbnail$suffix.js",
            ['jquery'],
            WC_STEP_FILTER_VERSION,
            true
        );

        wp_enqueue_script(
            'woocommerce-step-filter-multi-select',
            WC_STEP_FILTER_PLUGIN_URL . "$path/admin/js/multi-select$suffix.js",
            'jquery',
            WC_STEP_FILTER_PLUGIN_URL,
            true
        );

        wp_enqueue_script(
            'woocommerce-step-filter-hooks',
            WC_STEP_FILTER_PLUGIN_URL . "$path/admin/js/hooks$suffix.js",
            ['jquery'],
            WC_STEP_FILTER_VERSION,
            true
        );

        wp_enqueue_style(
            'woocommerce-step-filter',
            WC_STEP_FILTER_PLUGIN_URL . "$path/admin/$stylesFolder/app$suffix.css",
            [],
            WC_STEP_FILTER_VERSION
        );
    }

    /** Styles and scripts enqueue */
    public function enqueueFront()
    {
        $path = WC_STEP_FILTER_DEBUG ? 'src' : 'assets';
        $suffix = WC_STEP_FILTER_DEBUG ? '' : '.min';
        $stylesFolder = WC_STEP_FILTER_DEBUG ? 'scss' : 'css';
        $scriptsIncludingType = self::getSetting('scripts_including_type');
        $includedScripts = self::getSetting('included_scripts');

        wp_enqueue_script('jquery');

        if ($scriptsIncludingType == 'single' && !WC_STEP_FILTER_DEBUG) {
            wp_enqueue_script(
                'scripts',
                WC_STEP_FILTER_PLUGIN_URL . 'assets/front/js/scripts.min.js',
                ['jquery'],
                WC_STEP_FILTER_VERSION,
                true
            );
        } elseif ($scriptsIncludingType == 'multiple' || WC_STEP_FILTER_DEBUG) {
            if (in_array('jquery.serializejson', $includedScripts)) {
                wp_enqueue_script(
                    'jquery.serializejson',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/js/jquery.serializejson$suffix.js",
                    ['jquery'],
                    '2.9.0',
                    true
                );
            }

            if (in_array('wNumb', $includedScripts)) {
                wp_enqueue_script(
                    'wNumb',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/js/wNumb$suffix.js",
                    ['jquery'],
                    '1.1.0',
                    true
                );
            }

            if (in_array('nouislider', $includedScripts)) {
                wp_enqueue_script(
                    'nouislider',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/js/nouislider$suffix.js",
                    ['jquery', 'wNumb'],
                    '14.0.2',
                    true
                );
            }

            if (in_array('nouislider-launcher', $includedScripts)) {
                wp_enqueue_script(
                    'nouislider-launcher',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/js/nouislider-launcher$suffix.js",
                    ['jquery', 'nouislider'],
                    WC_STEP_FILTER_VERSION,
                    true
                );
            }

            if (in_array('app', $includedScripts)) {
                wp_enqueue_script(
                    'woocommerce-step-filter',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/js/app$suffix.js",
                    ['jquery'],
                    WC_STEP_FILTER_VERSION,
                    true
                );
            }

            if (in_array('hooks', $includedScripts)) {
                wp_enqueue_script(
                    'woocommerce-step-filter-hooks',
                    WC_STEP_FILTER_PLUGIN_URL . "$path/front/js/hooks$suffix.js",
                    ['jquery', 'woocommerce-step-filter'],
                    WC_STEP_FILTER_VERSION,
                    true
                );
            }
        }

        if (filter_var(self::getSetting('include_full_style'), FILTER_VALIDATE_BOOLEAN)) {
            wp_enqueue_style(
                'woocommerce-step-filter-full',
                WC_STEP_FILTER_PLUGIN_URL . "$path/front/$stylesFolder/app-full$suffix.css",
                [],
                WC_STEP_FILTER_VERSION
            );
        } else {
            wp_enqueue_style(
                'woocommerce-step-filter',
                WC_STEP_FILTER_PLUGIN_URL . "$path/front/$stylesFolder/app$suffix.css",
                [],
                WC_STEP_FILTER_VERSION
            );
        }
    }
    // </editor-fold>

    // <editor-fold desc="Settings">
    /**
     * Add the tab to the woocommerce options page
     *
     * @param array $settingsTabs
     *
     * @return array $settingsTabs
     */
    public function addSettingsTab($settingsTabs)
    {
        $settingsTabs['step_filter_settings_tab'] = 'Step Filter';

        return $settingsTabs;
    }

    /** Init options fields */
    public function initSettingsTab()
    {
        $defaults = [
            'section_title' => [
                'name' => 'Step Filter',
                'type' => 'title',
                'desc' => '',
                'id' => 'wc_step_filter_section_title'
            ]
        ];

        $args = array_merge($defaults, self::$settingsModel);

        woocommerce_admin_fields($args);
        woocommerce_admin_fields(
            [
                'section_end' => [
                    'type' => 'sectionend',
                    'id' => 'wc_step_filter_section_end'
                ]
            ]
        );
    }

    /** Save options fields */
    public function updateSettingsTab()
    {
        woocommerce_update_options(self::$settingsModel);
    }

    /**
     * Get one of settings
     *
     * @param string $setting
     *
     * @return string|boolean|array|float
     */
    public static function getSetting($setting)
    {
        if (!isset(self::$settingsModel[$setting])) {
            return null;
        }

        $value = get_option(self::$settingsModel[$setting]['id'], null);

        if (!is_null($value)) {
            return $value;
        }

        if (isset(self::$settingsModel[$setting]['default'])) {
            return self::$settingsModel[$setting]['default'];
        }

        return null;
    }

    /**
     * Generate html-field from passed args
     *
     * @param array $modelItem
     * @param array $args
     */
    public static function settingFieldView($modelItem, $args = [])
    {
        $defaultArgs = [
            'values' => [],
            'namePattern' => '%key%',
            'generateId' => true,
            'asTemplate' => false
        ];

        $args = array_replace_recursive($defaultArgs, $args);

        // create name from pattern
        $args['name'] = str_replace('%key%', $modelItem['key'], $args['namePattern']);

        // define value
        $args['value'] = isset($args['values'][$modelItem['key']])
            ? $args['values'][$modelItem['key']]
            : $modelItem['default'];

        // define id attribute
        $args['id'] = $args['generateId'] ? $modelItem['key'] : false;

        $viewPath = __DIR__ . DIRECTORY_SEPARATOR
            . 'setting-field-views' . DIRECTORY_SEPARATOR . $modelItem['type'] . '.php';

        echo isset($modelItem['before']) ? $modelItem['before'] : '';

        if (file_exists($viewPath)) {
            include($viewPath);
        }

        echo isset($modelItem['after']) ? $modelItem['after'] : '';

        if (isset($modelItem['description'])) {
            echo '<p class="description">' . wp_kses_post($modelItem['description']) . '</p>';
        }
    }
    // </editor-fold>

    // <editor-fold desc="Attribute">
    /**
     * Add metabox to attribute value edit page
     *
     * @param object $term
     */
    public function attributeValueFieldsAdd($term)
    {
        foreach (self::$attributeValueSettingsModel as $setting) {
            $value = get_term_meta($term->term_id, $setting['key']);
            ?>
            <tr valign="top">
                <th scope="row" class="titledesc">
                    <label for="<?php echo esc_attr($setting['key']); ?>"><?php
                    echo wp_kses_post($setting['label']);
                    ?></label>
                </th>
                <td><?php
                    self::settingFieldView(
                        $setting,
                        ['values' => [$setting['key'] => isset($value[0]) ? $value[0] : null]]
                    );
                    ?></td>
            </tr>
            <?php
        }
    }

    /**
     * Save attribute value meta fields
     *
     * @param integer $termId
     */
    public function attributeValueFieldsSave($termId)
    {
        if (!$termId) {
            return;
        }

        foreach (self::$attributeValueSettingsModel as $setting) {
            $value = '';

            if (isset($_POST[$setting['key']])) {
                $value = $_POST[$setting['key']];
            }

            update_term_meta($termId, $setting['key'], $value);
        }
    }
    // </editor-fold>

    // <editor-fold desc="Category">
    /**
     * Add metabox to category edit page
     *
     * @param object $term
     */
    public function categoryFieldsAdd($term)
    {
        foreach (self::$categorySettingsModel as $setting) {
            $value = get_term_meta($term->term_id, $setting['key']);
            ?>
            <tr valign="top">
                <th scope="row" class="titledesc">
                    <label for="<?php echo esc_attr($setting['key']); ?>"><?php
                        echo wp_kses_post($setting['label']);
                        ?></label>
                </th>
                <td><?php
                    self::settingFieldView(
                        $setting,
                        ['values' => [$setting['key'] => isset($value[0]) ? $value[0] : null]]
                    );
                    ?></td>
            </tr>
            <?php
        }
    }

    /**
     * Save tag meta fields
     *
     * @param integer $termId
     */
    public function categoryFieldsSave($termId)
    {
        if (!$termId) {
            return;
        }

        foreach (self::$categorySettingsModel as $setting) {
            $value = '';

            if (isset($_POST[$setting['key']])) {
                $value = $_POST[$setting['key']];
            }

            update_term_meta($termId, $setting['key'], $value);
        }
    }
    // </editor-fold>

    // <editor-fold desc="Utils">
    /**
     * Build URL from parts
     *
     * @param array $parts
     *
     * @return string
     */
    public static function buildUrl($parts)
    {
        $scheme = isset($parts['scheme']) ? ($parts['scheme'] . '://') : '';
        $host = isset($parts['host']) ? $parts['host'] : '';
        $port = isset($parts['port']) ? (':' . $parts['port']) : '';
        $user = isset($parts['user']) ? $parts['user'] : '';
        $pass = isset($parts['pass']) ? (':' . $parts['pass']) : '';
        $pass = ($user || $pass) ? ($pass . '@') : '';
        $path = isset($parts['path']) ? $parts['path'] : '';
        $query = isset($parts['query']) ? ('?' . $parts['query']) : '';
        $fragment = isset($parts['fragment']) ? ('#' . $parts['fragment']) : '';

        return implode('', [$scheme, $user, $pass, $host, $port, $path, $query, $fragment]);
    }

    /**
     * Get all unique products prices array
     *
     * @param array $ids - specific products ids or nothing
     *
     * @return array
     */
    public static function getProductsPrices($ids = [])
    {
        global $wpdb;

        $sql = "SELECT DISTINCT pm.meta_value FROM "
            . "{$wpdb->postmeta} as pm WHERE pm.meta_key = '_price' ORDER BY pm.meta_value + 0";

        if (!empty($ids)) {
            $inQuery = ' AND pm.post_id IN (' . implode(', ', array_fill(0, count($ids), '%s')) . ')';
            $sql .= call_user_func_array([$wpdb, 'prepare'], array_merge([$inQuery], $ids));
        }

        return $wpdb->get_col($sql);
    }

    /**
     * Get min and max prices of all products
     *
     * @param array $ids - specific products ids or nothing
     *
     * @return array
     */
    public static function getMinMaxPrices($ids = [])
    {
        global $wpdb;

        $sql = 'SELECT MIN(CAST(pm.meta_value as UNSIGNED)) as min, MAX(CAST(pm.meta_value as UNSIGNED)) as max FROM '
            . "{$wpdb->postmeta} as pm WHERE pm.meta_key = '_price'";

        if (!empty($ids)) {
            $inQuery = ' AND pm.post_id IN (' . implode(', ', array_fill(0, count($ids), '%s')) . ')';
            $sql .= call_user_func_array([$wpdb, 'prepare'], array_merge([$inQuery], $ids));
        }

        $results = $wpdb->get_results($sql);

        if (empty($results)) {
            return [
                'min' => null,
                'max' => null
            ];
        }

        $result = array_shift($results);

        return apply_filters(
            'wcsf_min_max_prices',
            [
                'min' => (float) $result->min,
                'max' => (float) $result->max
            ]
        );
    }

    /**
     * Get min and max numeric attribute values
     *
     * @param integer $id - attribute id
     * @param array $ids - specific products ids or nothing
     *
     * @return array
     */
    public static function getMinMaAttributeValues($id, $ids = [])
    {
        $attributeTaxonomy = wc_attribute_taxonomy_name_by_id((int) $id);
        $terms = get_terms($attributeTaxonomy);

        if (!empty($ids)) {
            // filter by products ids
            foreach ($terms as $key => $term) {
                $enabled = false;

                foreach ($ids as $productId) {
                    if (has_term($term->term_id, $attributeTaxonomy, $productId)) {
                        $enabled = true;

                        break;
                    }
                }

                if (!$enabled) {
                    unset($terms[$key]);
                }
            }
        }

        $values = array_map(function ($o) {
            return (float) $o->name;
        }, $terms);

        return apply_filters(
            'wcsf_min_max_attribute_values',
            [
                'min' => (float) min($values),
                'max' => (float) max($values)
            ],
            $id
        );
    }

    /**
     * Get min and max numeric meta values
     *
     * @param string $key
     * @param array $ids - specific products ids or nothing
     *
     * @return array
     */
    public static function getMinMaxMetaValues($key, $ids = [])
    {
        if (empty($key)) {
            return [
                'min' => null,
                'max' => null
            ];
        }

        global $wpdb;

        $sql = $wpdb->prepare(
            'SELECT MIN(CAST(pm.meta_value as UNSIGNED)) as min, MAX(CAST(pm.meta_value as UNSIGNED)) as max FROM '
            . "{$wpdb->postmeta} as pm WHERE pm.meta_key = '%s'",
            $key
        );

        if (!empty($ids)) {
            $inQuery = ' AND pm.post_id IN (' . implode(', ', array_fill(0, count($ids), '%s')) . ')';
            $sql .= call_user_func_array([$wpdb, 'prepare'], array_merge([$inQuery], $ids));
        }

        $results = $wpdb->get_results($sql);

        if (empty($results)) {
            return [
                'min' => null,
                'max' => null
            ];
        }

        $result = array_shift($results);

        return apply_filters(
            'wcsf_min_max_meta_values',
            [
                'min' => (float) $result->min,
                'max' => (float) $result->max
            ]
        );
    }

    /**
     * Get possible meta values array
     *
     * @param string $key
     *
     * @return array
     */
    public static function getMetaValues($key)
    {
        if (empty($key)) {
            return [];
        }

        global $wpdb;

        $output = [];
        $sql = $wpdb->prepare(
            'SELECT DISTINCT pm.post_id, pm.meta_value FROM '
            . "{$wpdb->postmeta} as pm WHERE pm.meta_key = '%s' ORDER BY pm.meta_value",
            $key
        );

        $results = $wpdb->get_results($sql);

        foreach ($results as $result) {
            $output[$result->post_id] = $result->meta_value;
        }

        asort($output);

        return apply_filters('wcsf_meta_values', $output, $key);
    }

    /**
     * Get multi-dimension array of the taxonomy terms
     *
     * @param array $queryArgs
     * @param array $args
     *
     * @return array
     */
    public static function getTermsTree($queryArgs = [], $args = [])
    {
        $output = [];
        $defaults = [
            'flatten' => false,
            'showLevelInName' => false,
            'level' => 0,
            'maxLevel' => 20
        ];

        $args = array_replace($defaults, $args);

        if ($args['level'] > $args['maxLevel']) {
            return $output;
        }

        $defaultQuery = [
            'taxonomy' => 'product_cat',
            'hide_empty' => false,
            'hierarchical' => false,
            'parent' => 0
        ];

        $queryArgs = array_replace($defaultQuery, $queryArgs);
        $terms = get_terms($queryArgs);

        foreach ($terms as $term) {
            $children = self::getTermsTree(
                array_replace($queryArgs, ['parent' => $term->term_id]),
                array_replace($args, ['level' => $args['level'] + 1])
            );

            if ($args['showLevelInName']) {
                $term->name = str_repeat('—', $args['level']) . $term->name;
            }

            if ($args['flatten']) {
                $output[$term->term_id] = $term;
                $output = array_merge($output, $children);

                continue;
            }

            $output[$term->term_id] = $term;

            if (!empty($children)) {
                $term->children = $children;
            }
        }

        return $output;
    }
    // </editor-fold>
}
