/* wcStepFilterQuestion
 * Original author: troll_winner@mail.ru
 * Further changes, comments: troll_winner@mail.ru
 */

(function (root, factory) {
    'use strict';

    if (typeof define === 'function' && define.amd) {
        define(['jquery'], factory);
    } else if (typeof exports === 'object'
        && typeof module !== 'undefined'
        && typeof require === 'function'
    ) {
        module.exports = factory(require('jquery'));
    } else {
        factory(root.jQuery);
    }
})(this, function ($) {
    'use strict';

    const pluginName = 'wcStepFilterQuestion';
    const defaults = {};

    const Plugin = function (element, options) {
        this.element = element;
        this.options = $.extend({}, defaults, options);

        return this.init();
    };

    /**
     * Init the instance
     * @returns {Object} self instance
     */
    Plugin.prototype.init = function () {
        const _this = this;

        this.$element = $(this.element);
        this.$settings = this.$element.find('[data-component="wcsf-question-setting"]');
        this.$manualValues = this.$element.find('[data-component="wcsf-question-manual-values"]');
        this.$filterType = this.$element.find('#_filter_type');
        this.$filterTypes = this.$filterType.find('option');
        this.$viewType = this.$element.find('#_view_type');
        this.ajaxUrl = this.$element.data('ajax-url');

        this.toggleFields();

        this.$element
            .on('change', '#_filter_type', () => this.toggleFields())
            .on('change', '#_view_type', () => this.toggleFields())

            // add manual filter row
            .on('click', '[data-component="wcsf-question-add-value"]', function (event) {
                event.preventDefault();

                const $element = $(this);

                $element.attr('disabled', true);

                return _this
                    .addValue()
                    .then(() => $element.attr('disabled', false));
            })

            // remove manual filter row
            .on('click', '[data-component="wcsf-question-manual-value-remove"]', function (event) {
                event.preventDefault();

                const $element = $(this);
                const $value = _this.$element.find('[data-component="wcsf-question-manual-value"]');
                const $valueToRemove = $value.filter(`[data-index="${$element.data('index')}"]`);

                if ($value.length !== 1) {
                    $valueToRemove.remove();

                    _this.recalculate();
                }
            });

        return this;
    };

    /**
     * Show/hide fields according the type
     * @returns {Object} self instance
     */
    Plugin.prototype.toggleFields = function () {
        let filterType = this.$filterType.val();
        const viewType = this.$viewType.val();
        const typesVisibilityRules = {
            price: ['number-from', 'number-to', 'number-between'],
            manual: ['radio', 'select', 'checkbox', 'image-radio', 'image-checkbox'],
            tag: [
                'radio',
                'select',
                'checkbox',
                'image-radio',
                'image-checkbox',
                'multi-choice-checkbox',
                'multi-choice-range'
            ],
            category: [
                'radio',
                'select',
                'checkbox',
                'image-radio',
                'image-checkbox',
                'multi-choice-checkbox',
                'multi-choice-range'
            ],
            meta: [
                'radio',
                'select',
                'checkbox',
                'multi-choice-checkbox',
                'multi-choice-range',
                'number-from',
                'number-to',
                'number-between'
            ]
        };

        const settingsVisibilityRulesByView = {
            from_value_text: ['number-from', 'number-to', 'number-between'],
            to_value_text: ['number-from', 'number-to', 'number-between'],
            numeric_unit: ['number-from', 'number-to', 'number-between'],
            numeric_step: ['number-from', 'number-to', 'number-between'],
            any_product_value: ['checkbox', 'radio', 'select', 'image-radio'],
            any_product_value_text: ['checkbox', 'radio', 'select', 'image-radio'],
            any_product_value_thumbnail: ['checkbox', 'radio', 'select', 'image-radio'],
            alignment: ['image-radio', 'image-checkbox'],
            thumbnail_size: ['image-radio', 'image-checkbox'],
            multi_choice_values: ['multi-choice-checkbox', 'multi-choice-range'],
            multi_choice_undefined_value_text: ['multi-choice-checkbox', 'multi-choice-range'],
            multi_choice_possible_value_text: ['multi-choice-checkbox', 'multi-choice-range'],
            multi_choice_required_value_text: ['multi-choice-checkbox', 'multi-choice-range'],
            multi_choice_excluded_value_text: ['multi-choice-checkbox', 'multi-choice-range'],
            disable_unmet_values: [
                'select',
                'radio',
                'checkbox',
                'image-radio',
                'image-checkbox',
                'number-from',
                'number-to',
                'number-between'
            ],
            show_count: [
                'select',
                'radio',
                'checkbox',
                'image-radio',
                'image-checkbox',
                'multi-choice-checkbox',
                'multi-choice-range'
            ],
            action_on_click: [
                'select',
                'radio',
                'checkbox',
                'image-radio',
                'image-checkbox',
                'multi-choice-checkbox',
                'multi-choice-range'
            ],
            relation: ['checkbox', 'image-checkbox']
        };

        const settingsVisibilityRulesByType = {
            attribute: ['attribute'],
            any_product_value: ['attribute', 'tag', 'category'],
            any_product_value_text: ['attribute', 'tag', 'category'],
            any_product_value_thumbnail: ['attribute', 'tag', 'category'],
            category_parent: ['category'],
            include_sub_categories: ['category'],
            included_terms: ['attribute', 'tag', 'category'],
            excluded_terms: ['attribute', 'tag', 'category'],
            show_count: ['attribute', 'tag', 'category'],
            manual: ['manual'],
            meta_key: ['meta']
        };

        // toggle types according to the selected view
        this.$filterTypes.each(function () {
            const $element = $(this);

            if (!typesVisibilityRules.hasOwnProperty($element.attr('value'))) {
                return null;
            }

            const hide = $.inArray(viewType, typesVisibilityRules[$element.attr('value')]) === -1;

            $element.attr('disabled', hide);

            return this;
        });

        // reset select if disabled option was selected
        if (!this.$filterType.val()) {
            this.$filterType.val(this.$filterType.find('option:not(disabled):eq(0)').attr('value'));
            filterType = this.$filterType.val();
        }

        // toggle settings according to the selected view
        this.$settings.each(function () {
            const $element = $(this);

            if (!settingsVisibilityRulesByView.hasOwnProperty($element.data('key'))) {
                return null;
            }

            const hide = $.inArray(viewType, settingsVisibilityRulesByView[$element.data('key')]) === -1;

            $element[hide ? 'addClass' : 'removeClass']('is-hidden-by-view');

            return this;
        });

        // toggle settings according to the selected type
        this.$settings.each(function () {
            const $element = $(this);

            if (!settingsVisibilityRulesByType.hasOwnProperty($element.data('key'))) {
                return null;
            }

            const hide = $.inArray(filterType, settingsVisibilityRulesByType[$element.data('key')]) === -1;

            $element[hide ? 'addClass' : 'removeClass']('is-hidden-by-type');

            return this;
        });

        return this;
    };

    /**
     * Makes the ajax-request and add manual value fields row
     * @returns {Promise} ajax request
     */
    Plugin.prototype.addValue = function () {
        const data = {
            action: 'wcsfAddQuestionValue',
            postId: this.$element.data('post-id'),
            index: this.$manualValues.find('[data-component="wcsf-question-manual-value"]').length
        };

        return $.get(
            this.ajaxUrl,
            data,
            (response) => {
                this.$manualValues.append(response);

                $(document).trigger('addQuestionValue.wcStepFilterQuestion');
            }
        );
    };

    /**
     * ReCalculate input names indexes
     * @return {Object} jQuery elements
     */
    Plugin.prototype.recalculate = function () {
        return this.$element
            .find('[data-component="wcsf-question-manual-value"]')
            .each(function (index) {
                const $element = $(this);

                // change item index
                $element.attr('data-index', index);

                // change item children index
                $element
                    .find('[data-index]')
                    .each(function () {
                        return $(this).attr('data-index', index);
                    });

                return $element
                    .find(':input')
                    .each(function () {
                        const $input = $(this);
                        const name = $input.attr('name');

                        if (!name) {
                            return this;
                        }

                        const number = index.toString().split('').reverse().join('');

                        // replace the first array key from the end
                        return $input.attr(
                            'name',
                            name
                                .split('')
                                .reverse()
                                .join('')
                                .replace(/]\d+\[/, `]${number}[`)
                                .split('')
                                .reverse()
                                .join('')
                        );
                    });
            });
    };

    $.fn[pluginName] = function (options) {
        return this.each(function () {
            if (!$.data(this, 'plugin_' + pluginName)) {
                $.data(this, 'plugin_' + pluginName, new Plugin(this, options));
            }
        });
    };

    const init = () => $('[data-component="wcsf-question"]').each(function () {
        return $(this).wcStepFilterQuestion();
    });

    $(document)
        .ready(() => init())
        .on('init.wcStepFilterQuestion', () => init())

        // toggle manual filter source sections
        .on('change', '[data-component="wcsf-question-manual-value-filter-by"]', function () {
            return $(this)
                .closest('[data-component="wcsf-modal"]')
                .find('[data-component="wcsf-question-manual-value-filter"]')
                .addClass('is-closed')
                .filter(`[data-id="${$(this).data('id')}"]`)
                .removeClass('is-closed');
        });
});
