/* Woocommerce Step Filter
 * Original author: troll_winner@mail.ru
 * Further changes, comments: troll_winner@mail.ru
 */

(function (root, factory) {
    'use strict';

    if (typeof define === 'function' && define.amd) {
        define(['jquery', 'jquery.serializejson'], factory);
    } else if (typeof exports === 'object'
        && typeof module !== 'undefined'
        && typeof require === 'function'
    ) {
        module.exports = factory(require('jquery'), require('jquery.serializejson'));
    } else {
        factory(root.jQuery);
    }
})(this, function ($) {
    'use strict';

    const pluginName = 'wcsf';
    const defaults = {
        documentNode: document,
        windowNode: window,
        bodySelector: 'html, body',
        preliminaryResultsCountRequestInterval: 1000
    };

    const Plugin = function (element, options) {
        this.element = element;
        this.options = $.extend({}, defaults, options);

        return this.init();
    };

    /**
     * Init the instance
     * @returns {Object} self instance
     */
    Plugin.prototype.init = function () {
        this.$element = $(this.element);
        this.$document = $(this.options.documentNode);
        this.$window = $(this.options.windowNode);
        this.$body = $(this.options.bodySelector);
        this.ajaxUrl = this.$element.data('ajax-url');

        // save instance as element data
        this.$element.data('wcsf', this);

        return this;
    };

    /**
     * Makes the ajax-request
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    Plugin.prototype.ajaxRequest = function (args) {
        args.id = this.$element.data('id');
        args.referer = this.options.documentNode.location.href;

        this.$element.addClass('is-loading');

        return $.post(
            this.ajaxUrl,
            args,
            (response) => {
                // make redirect only
                if (response.redirect) {
                    this.options.documentNode.location = response.redirect;

                    return this;
                }

                this.$element.html(response.content).removeClass('is-loading');

                // scroll to top
                if (this.$element.data('settings').scrolling_top_on_update) {
                    let topPosition = this.$element.offset().top;

                    if (this.$window.scrollTop() > topPosition) {
                        if (this.$element.data('settings').scrolling_up_gap) {
                            topPosition -= Number(this.$element.data('settings').scrolling_up_gap);
                        }

                        this.$body.stop().animate({scrollTop: topPosition}, 500);
                    }
                }

                this.$document.trigger('ajaxCompleted.wcsf', [this, response]);

                return this;
            },
            'json'
        );
    };

    /**
     * Get current step filter value
     * @returns {Array} value
     */
    Plugin.prototype.getValue = function () {
        const $form = this.$element.find('[data-component="wcsf-form"]');
        const value = $form.serializeJSON();

        return value.value;
    };

    /**
     * Submit step and go next
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    Plugin.prototype.submitAndGoNext = function (args) {
        args.stepId = args.stepId || this.$element.find('[data-component="wcsf-form"]').data('step-id');
        args.action = 'WCSFSubmitAndGoNext';
        args.value = this.getValue();

        return this.ajaxRequest(args);
    };

    /**
     * Submit step
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    Plugin.prototype.submit = function (args) {
        args.stepId = args.stepId || this.$element.find('[data-component="wcsf-form"]').data('step-id');
        args.action = 'WCSFSubmit';
        args.value = this.getValue();

        return this.ajaxRequest(args);
    };

    /**
     * Submit step and get another one by id
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    Plugin.prototype.submitAndGoTo = function (args) {
        args.action = 'WCSFGetStep';
        args.value = this.getValue();

        return this.ajaxRequest(args);
    };

    /**
     * Get step by id
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    Plugin.prototype.goTo = function (args) {
        args.action = 'WCSFGetStep';

        return this.ajaxRequest(args);
    };

    /**
     * Get step by id
     * @param {Object} $element - caller
     * @return {Object} self instance
     */
    Plugin.prototype.getPreliminaryResultsCount = function ($element) {
        const args = {
            action: 'WCSFGetPreliminaryResultsCount',
            value: this.getValue(),
            id: this.$element.data('id'),
            categoryId: this.$element.data('category-id'),
            tagId: this.$element.data('tag-id')
        };

        clearTimeout(this.getPreliminaryResultsCountTimeout);

        // pass cart to the server
        this.getPreliminaryResultsCountTimeout = setTimeout(
            () => $.post(
                this.ajaxUrl,
                args,
                (response) => this.$document.trigger('getPreliminaryResultsCount.wcsf', [this, response, $element]),
                'json'
            ),
            this.options.preliminaryResultsCountRequestInterval
        );

        return this;
    };

    /**
     * Reset form to the first step
     * @param {Object} args - object of arguments
     * @returns {Promise} ajax request
     */
    Plugin.prototype.reset = function (args = {}) {
        args.action = 'WCSFReset';

        return this.ajaxRequest(args);
    };

    $.fn[pluginName] = function (options) {
        return this.each(function () {
            if (!$.data(this, 'plugin_' + pluginName)) {
                $.data(this, 'plugin_' + pluginName, new Plugin(this, options));
            }
        });
    };
});
